@@ -12,7 +12,7 @@
 #include <stdint.h>
 #include <ctype.h>
 
-#include <libmount.h>
+#include <libsmartcols.h>
 
 #include "c.h"
 #include "nls.h"
@@ -20,12 +20,13 @@
 #include "blkdev.h"
 #include "pathnames.h"
 #include "xalloc.h"
+#include "strutils.h"
+#include "optutils.h"
 #include "closestream.h"
 
 #include "swapheader.h"
+#include "swapprober.h"
 #include "swapon-common.h"
-#include "strutils.h"
-#include "tt.h"
 
 #define PATH_MKSWAP	"/sbin/mkswap"
 
@@ -76,9 +77,6 @@ enum {
 	SIG_SWSUSPEND
 };
 
-#define SWAP_SIGNATURE		"SWAPSPACE2"
-#define SWAP_SIGNATURE_SZ	(sizeof(SWAP_SIGNATURE) - 1)
-
 static int all;
 static int priority = -1;	/* non-prioritized swap by default */
 static int discard;		/* don't send swap discards by default */
@@ -92,19 +90,35 @@ static int verbose;
 struct colinfo {
         const char *name; /* header */
         double     whint; /* width hint (N < 1 is in percent of termwidth) */
-        int        flags; /* TT_FL_* */
+	int        flags; /* SCOLS_FL_* */
         const char *help;
 };
-enum { COL_PATH, COL_TYPE, COL_SIZE, COL_USED, COL_PRIO };
+
+/* basic output flags */
+static int no_headings;
+static int raw;
+
+enum {
+	COL_PATH,
+	COL_TYPE,
+	COL_SIZE,
+	COL_USED,
+	COL_PRIO,
+	COL_UUID,
+	COL_LABEL
+};
 struct colinfo infos[] = {
 	[COL_PATH]     = { "NAME",	0.20, 0, N_("device file or partition path") },
-	[COL_TYPE]     = { "TYPE",	0.20, TT_FL_TRUNC, N_("type of the device")},
-	[COL_SIZE]     = { "SIZE",	0.20, TT_FL_RIGHT, N_("size of the swap area")},
-	[COL_USED]     = { "USED",	0.20, TT_FL_RIGHT, N_("bytes in use")},
-	[COL_PRIO]     = { "PRIO",	0.20, TT_FL_RIGHT, N_("swap priority")},
+	[COL_TYPE]     = { "TYPE",	0.20, SCOLS_FL_TRUNC, N_("type of the device")},
+	[COL_SIZE]     = { "SIZE",	0.20, SCOLS_FL_RIGHT, N_("size of the swap area")},
+	[COL_USED]     = { "USED",	0.20, SCOLS_FL_RIGHT, N_("bytes in use")},
+	[COL_PRIO]     = { "PRIO",	0.20, SCOLS_FL_RIGHT, N_("swap priority")},
+	[COL_UUID]     = { "UUID",	0.20, 0, N_("swap uuid")},
+	[COL_LABEL]    = { "LABEL",	0.20, 0, N_("swap label")},
 };
-#define NCOLS ARRAY_SIZE(infos)
-static int columns[NCOLS], ncolumns;
+
+static int columns[ARRAY_SIZE(infos) * 2];
+static int ncolumns;
 
 static int column_name_to_id(const char *name, size_t namesz)
 {
@@ -112,7 +126,7 @@ static int column_name_to_id(const char *name, size_t namesz)
 
 	assert(name);
 
-	for (i = 0; i < NCOLS; i++) {
+	for (i = 0; i < ARRAY_SIZE(infos); i++) {
 		const char *cn = infos[i].name;
 
 		if (!strncasecmp(name, cn, namesz) && !*(cn + namesz))
@@ -124,9 +138,8 @@ static int column_name_to_id(const char *name, size_t namesz)
 
 static inline int get_column_id(int num)
 {
-	assert(ARRAY_SIZE(columns) == NCOLS);
 	assert(num < ncolumns);
-	assert(columns[num] < (int)NCOLS);
+	assert(columns[num] < (int) ARRAY_SIZE(infos));
 
 	return columns[num];
 }
@@ -136,20 +149,22 @@ static inline struct colinfo *get_column_info(unsigned num)
 	return &infos[get_column_id(num)];
 }
 
-static void add_tt_line(struct tt *tt, struct libmnt_fs *fs, int bytes)
+static void add_scols_line(struct libscols_table *table, struct libmnt_fs *fs, int bytes)
 {
 	int i;
-	struct tt_line *line;
+	struct libscols_line *line;
+	blkid_probe pr = NULL;
+	const char *data;
 
-	assert(tt);
+	assert(table);
 	assert(fs);
 
-	line = tt_add_line(tt, NULL);
-	if (!line) {
-		warn(_("failed to add line to output"));
-		return;
-	}
-
+	line = scols_table_new_line(table, NULL);
+	if (!line)
+		err(EXIT_FAILURE, _("failed to initialize output line"));
+	data = mnt_fs_get_source(fs);
+	if (access(data, R_OK) == 0)
+		pr = get_swap_prober(data);
 	for (i = 0; i < ncolumns; i++) {
 		char *str = NULL;
 		off_t size;
@@ -180,13 +195,23 @@ static void add_tt_line(struct tt *tt, struct libmnt_fs *fs, int bytes)
 		case COL_PRIO:
 			xasprintf(&str, "%d", mnt_fs_get_priority(fs));
 			break;
+		case COL_UUID:
+			if (pr && !blkid_probe_lookup_value(pr, "UUID", &data, NULL))
+				xasprintf(&str, "%s", data);
+			break;
+		case COL_LABEL:
+			if (pr && !blkid_probe_lookup_value(pr, "LABEL", &data, NULL))
+				xasprintf(&str, "%s", data);
+			break;
 		default:
 			break;
 		}
 
 		if (str)
-			tt_line_set_data(line, i, str);
+			scols_line_refer_data(line, i, str);
 	}
+	if (pr)
+		blkid_free_probe(pr);
 	return;
 }
 
@@ -209,7 +234,7 @@ static int display_summary(void)
 	printf(_("%s\t\t\t\tType\t\tSize\tUsed\tPriority\n"), _("Filename"));
 
 	while (mnt_table_next_fs(st, itr, &fs) == 0) {
-		printf("%-39s\t%s\t%jd\t%jd\t%d\n",
+		printf("%-39s\t%-8s\t%jd\t%jd\t%d\n",
 			mnt_fs_get_source(fs),
 			mnt_fs_get_swaptype(fs),
 			mnt_fs_get_size(fs),
@@ -221,14 +246,13 @@ static int display_summary(void)
 	return 0;
 }
 
-static int show_table(int tt_flags, int bytes)
+static int show_table(int bytes)
 {
 	struct libmnt_table *st = get_swaps();
 	struct libmnt_iter *itr = NULL;
 	struct libmnt_fs *fs;
-
-	int i, rc = 0;
-	struct tt *tt = NULL;
+	int i;
+	struct libscols_table *table = NULL;
 
 	if (!st)
 		return -1;
@@ -237,30 +261,29 @@ static int show_table(int tt_flags, int bytes)
 	if (!itr)
 		err(EXIT_FAILURE, _("failed to initialize libmount iterator"));
 
-	tt = tt_new_table(tt_flags | TT_FL_FREEDATA);
-	if (!tt) {
-		warn(_("failed to initialize output table"));
-		goto done;
-	}
+	scols_init_debug(0);
+
+	table = scols_new_table();
+	if (!table)
+		err(EXIT_FAILURE, _("failed to initialize output table"));
+
+	scols_table_enable_raw(table, raw);
+	scols_table_enable_noheadings(table, no_headings);
 
 	for (i = 0; i < ncolumns; i++) {
 		struct colinfo *col = get_column_info(i);
 
-		if (!tt_define_column(tt, col->name, col->whint, col->flags)) {
-			warnx(_("failed to initialize output column"));
-			rc = -1;
-			goto done;
-		}
+		if (!scols_table_new_column(table, col->name, col->whint, col->flags))
+			err(EXIT_FAILURE, _("failed to initialize output column"));
 	}
 
 	while (mnt_table_next_fs(st, itr, &fs) == 0)
-		add_tt_line(tt, fs, bytes);
+		add_scols_line(table, fs, bytes);
 
-	tt_print_table(tt);
- done:
+	scols_print_table(table);
+	scols_unref_table(table);
 	mnt_free_iter(itr);
-	tt_free_table(tt);
-	return rc;
+	return 0;
 }
 
 /* calls mkswap */
@@ -345,8 +368,7 @@ err:
 
 static int swap_detect_signature(const char *buf, int *sig)
 {
-	if (memcmp(buf, "SWAP-SPACE", 10) == 0 ||
-            memcmp(buf, "SWAPSPACE2", 10) == 0)
+	if (memcmp(buf, SWAP_SIGNATURE, SWAP_SIGNATURE_SZ) == 0)
 		*sig = SIG_SWAPSPACE;
 
 	else if (memcmp(buf, "S1SUSPEND", 9) == 0 ||
@@ -403,21 +425,19 @@ static unsigned long long swap_get_size(const char *hdr, const char *devname,
 					unsigned int pagesize)
 {
 	unsigned int last_page = 0;
-	int swap_version = 0;
+	const unsigned int swap_version = SWAP_VERSION;
 	int flip = 0;
 	struct swap_header_v1_2 *s;
 
 	s = (struct swap_header_v1_2 *) hdr;
-	if (s->version == 1) {
-		swap_version = 1;
+	if (s->version == swap_version) {
 		last_page = s->last_page;
-	} else if (swab32(s->version) == 1) {
+	} else if (swab32(s->version) == swap_version) {
 		flip = 1;
-		swap_version = 1;
 		last_page = swab32(s->last_page);
 	}
 	if (verbose)
-		warnx(_("%s: found swap signature: version %d, "
+		warnx(_("%s: found swap signature: version %ud, "
 			"page-size %d, %s byte order"),
 			devname,
 			swap_version,
@@ -459,7 +479,7 @@ static int swapon_checks(const char *special)
 	int permMask;
 
 	if (stat(special, &st) < 0) {
-		warn(_("stat failed %s"), special);
+		warn(_("stat of %s failed"), special);
 		goto err;
 	}
 
@@ -490,13 +510,13 @@ static int swapon_checks(const char *special)
 	}
 
 	if (S_ISBLK(st.st_mode) && blkdev_get_size(fd, &devsize)) {
-		warn(_("%s: get size failed"), special);
+		warnx(_("%s: get size failed"), special);
 		goto err;
 	}
 
 	hdr = swap_get_header(fd, &sig, &pagesize);
 	if (!hdr) {
-		warn(_("%s: read swap header failed"), special);
+		warnx(_("%s: read swap header failed"), special);
 		goto err;
 	}
 
@@ -621,6 +641,42 @@ static int swapon_by_uuid(const char *uuid, int prio, int dsc)
 			 cannot_find(uuid);
 }
 
+/* -o <options> or fstab */
+static int parse_options(const char *optstr,
+			 int *prio, int *disc, int *nofail)
+{
+	char *arg = NULL;
+
+	assert(optstr);
+	assert(prio);
+	assert(disc);
+	assert(nofail);
+
+	if (mnt_optstr_get_option(optstr, "nofail", NULL, 0) == 0)
+		*nofail = 1;
+
+	if (mnt_optstr_get_option(optstr, "discard", &arg, NULL) == 0) {
+		*disc |= SWAP_FLAG_DISCARD;
+
+		if (arg) {
+			/* only single-time discards are wanted */
+			if (strcmp(arg, "once") == 0)
+				*disc |= SWAP_FLAG_DISCARD_ONCE;
+
+			/* do discard for every released swap page */
+			if (strcmp(arg, "pages") == 0)
+				*disc |= SWAP_FLAG_DISCARD_PAGES;
+			}
+	}
+
+	arg = NULL;
+	if (mnt_optstr_get_option(optstr, "pri", &arg, NULL) == 0 && arg)
+		*prio = atoi(arg);
+
+	return 0;
+}
+
+
 static int swapon_all(void)
 {
 	struct libmnt_table *tb = get_fstab();
@@ -638,26 +694,14 @@ static int swapon_all(void)
 	while (mnt_table_find_next_fs(tb, itr, match_swap, NULL, &fs) == 0) {
 		/* defaults */
 		int pri = priority, dsc = discard, nofail = ifexists;
-		char *p, *src, *dscarg;
+		const char *opts, *src;
 
 		if (mnt_fs_get_option(fs, "noauto", NULL, NULL) == 0)
 			continue;
-		if (mnt_fs_get_option(fs, "discard", &dscarg, NULL) == 0) {
-			dsc |= SWAP_FLAG_DISCARD;
-			if (dscarg) {
-				/* only single-time discards are wanted */
-				if (strcmp(dscarg, "once") == 0)
-					dsc |= SWAP_FLAG_DISCARD_ONCE;
 
-				/* do discard for every released swap page */
-				if (strcmp(dscarg, "pages") == 0)
-					dsc |= SWAP_FLAG_DISCARD_PAGES;
-			}
-		}
-		if (mnt_fs_get_option(fs, "nofail", NULL, NULL) == 0)
-			nofail = 1;
-		if (mnt_fs_get_option(fs, "pri", &p, NULL) == 0 && p)
-			pri = atoi(p);
+		opts = mnt_fs_get_options(fs);
+		if (opts)
+			parse_options(opts, &pri, &dsc, &nofail);
 
 		src = mnt_resolve_spec(mnt_fs_get_source(fs), mntcache);
 		if (!src) {
@@ -675,25 +719,29 @@ static int swapon_all(void)
 	return status;
 }
 
+
 static void __attribute__ ((__noreturn__)) usage(FILE * out)
 {
 	size_t i;
 	fputs(USAGE_HEADER, out);
-
 	fprintf(out, _(" %s [options] [<spec>]\n"), program_invocation_short_name);
 
+	fputs(USAGE_SEPARATOR, out);
+	fputs(_("Enable devices and files for paging and swapping.\n"), out);
+
 	fputs(USAGE_OPTIONS, out);
-	fputs(_(" -a, --all                enable all swaps from /etc/fstab\n"
-		" -d, --discard[=<policy>] enable swap discards, if supported by device\n"
-		" -e, --ifexists           silently skip devices that do not exist\n"
-		" -f, --fixpgsz            reinitialize the swap space if necessary\n"
-		" -p, --priority <prio>    specify the priority of the swap device\n"
-		" -s, --summary            display summary about used swap devices\n"
-		"     --show[=<columns>]   display summary in definable table\n"
-		"     --noheadings         don't print headings, use with --show\n"
-		"     --raw                use the raw output format, use with --show\n"
-		"     --bytes              display swap size in bytes in --show output\n"
-		" -v, --verbose            verbose mode\n"), out);
+	fputs(_(" -a, --all                enable all swaps from /etc/fstab\n"), out);
+	fputs(_(" -d, --discard[=<policy>] enable swap discards, if supported by device\n"), out);
+	fputs(_(" -e, --ifexists           silently skip devices that do not exist\n"), out);
+	fputs(_(" -f, --fixpgsz            reinitialize the swap space if necessary\n"), out);
+	fputs(_(" -o, --options <list>     comma-separated list of swap options\n"), out);
+	fputs(_(" -p, --priority <prio>    specify the priority of the swap device\n"), out);
+	fputs(_(" -s, --summary            display summary about used swap devices (DEPRECATED)\n"), out);
+	fputs(_("     --show[=<columns>]   display summary in definable table\n"), out);
+	fputs(_("     --noheadings         don't print table heading (with --show)\n"), out);
+	fputs(_("     --raw                use the raw output format (with --show)\n"), out);
+	fputs(_("     --bytes              display swap size in bytes in --show output\n"), out);
+	fputs(_(" -v, --verbose            verbose mode\n"), out);
 
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
@@ -710,13 +758,13 @@ static void __attribute__ ((__noreturn__)) usage(FILE * out)
 		" <file>                 name of file to be used\n"), out);
 
 	fputs(_("\nAvailable discard policy types (for --discard):\n"
-		" once	  : only single-time area discards are issued. (swapon)\n"
-		" pages	  : discard freed pages before they are reused.\n"
-		" * if no policy is selected both discard types are enabled. (default)\n"), out);
+		" once    : only single-time area discards are issued\n"
+		" pages   : freed pages are discarded before they are reused\n"
+		"If no policy is selected, both discard types are enabled (default).\n"), out);
 
 	fputs(_("\nAvailable columns (for --show):\n"), out);
-	for (i = 0; i < NCOLS; i++)
-		fprintf(out, " %4s  %s\n", infos[i].name, _(infos[i].help));
+	for (i = 0; i < ARRAY_SIZE(infos); i++)
+		fprintf(out, " %-5s  %s\n", infos[i].name, _(infos[i].help));
 
 	fprintf(out, USAGE_MAN_TAIL("swapon(8)"));
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
@@ -724,22 +772,24 @@ static void __attribute__ ((__noreturn__)) usage(FILE * out)
 
 int main(int argc, char *argv[])
 {
+	char *options = NULL;
 	int status = 0, c;
-	int show = 0, tt_flags = 0;
+	int show = 0;
 	int bytes = 0;
 	size_t i;
 
 	enum {
-		SHOW_OPTION = CHAR_MAX + 1,
-		RAW_OPTION,
+		BYTES_OPTION = CHAR_MAX + 1,
 		NOHEADINGS_OPTION,
-		BYTES_OPTION
+		RAW_OPTION,
+		SHOW_OPTION
 	};
 
 	static const struct option long_opts[] = {
 		{ "priority", 1, 0, 'p' },
 		{ "discard",  2, 0, 'd' },
 		{ "ifexists", 0, 0, 'e' },
+		{ "options",  2, 0, 'o' },
 		{ "summary",  0, 0, 's' },
 		{ "fixpgsz",  0, 0, 'f' },
 		{ "all",      0, 0, 'a' },
@@ -753,6 +803,15 @@ int main(int argc, char *argv[])
 		{ NULL, 0, 0, 0 }
 	};
 
+	static const ul_excl_t excl[] = {       /* rows and cols in in ASCII order */
+		{ 'a','o','s', SHOW_OPTION },
+		{ 'a','o', BYTES_OPTION },
+		{ 'a','o', NOHEADINGS_OPTION },
+		{ 'a','o', RAW_OPTION },
+		{ 0 }
+	};
+	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
+
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
@@ -761,8 +820,11 @@ int main(int argc, char *argv[])
 	mnt_init_debug(0);
 	mntcache = mnt_new_cache();
 
-	while ((c = getopt_long(argc, argv, "ahd::efp:svVL:U:",
+	while ((c = getopt_long(argc, argv, "ahd::efo:p:svVL:U:",
 				long_opts, NULL)) != -1) {
+
+		err_exclusive_options(c, long_opts, excl, excl_st);
+
 		switch (c) {
 		case 'a':		/* all */
 			++all;
@@ -770,6 +832,9 @@ int main(int argc, char *argv[])
 		case 'h':		/* help */
 			usage(stdout);
 			break;
+		case 'o':
+			options = optarg;
+			break;
 		case 'p':		/* priority */
 			priority = strtos16_or_err(optarg,
 					   _("failed to parse priority"));
@@ -818,10 +883,10 @@ int main(int argc, char *argv[])
 			show = 1;
 			break;
 		case NOHEADINGS_OPTION:
-			tt_flags |= TT_FL_NOHEADINGS;
+			no_headings = 1;
 			break;
 		case RAW_OPTION:
-			tt_flags |= TT_FL_RAW;
+			raw = 1;
 			break;
 		case BYTES_OPTION:
 			bytes = 1;
@@ -847,7 +912,7 @@ int main(int argc, char *argv[])
 			columns[ncolumns++] = COL_USED;
 			columns[ncolumns++] = COL_PRIO;
 		}
-		status = show_table(tt_flags, bytes);
+		status = show_table(bytes);
 		return status;
 	}
 
@@ -857,6 +922,9 @@ int main(int argc, char *argv[])
 	if (all)
 		status |= swapon_all();
 
+	if (options)
+		parse_options(options, &priority, &discard, &ifexists);
+
 	for (i = 0; i < numof_labels(); i++)
 		status |= swapon_by_label(get_label(i), priority, discard);
 
