@@ -68,6 +68,10 @@ function ts_check_losetup {
 	local tmp
 	ts_check_test_command "$TS_CMD_LOSETUP"
 
+	if [ "$TS_SKIP_LOOPDEVS" = "yes" ]; then
+		ts_skip "loop-device tests disabled"
+	fi
+
 	# assuming that losetup -f works ... to be checked somewhere else
 	tmp=$($TS_CMD_LOSETUP -f 2>/dev/null)
 	if test -b "$tmp"; then
@@ -82,9 +86,8 @@ function ts_skip_subtest {
 
 function ts_skip {
 	ts_skip_subtest "$1"
-	if [ -n "$2" -a -b "$2" ]; then
-		ts_device_deinit "$2"
-	fi
+
+	ts_cleanup_on_exit
 	exit 0
 }
 
@@ -237,6 +240,7 @@ function ts_init_env {
 	TS_VERBOSE=$(ts_has_option "verbose" "$*")
 	TS_PARALLEL=$(ts_has_option "parallel" "$*")
 	TS_KNOWN_FAIL=$(ts_has_option "known-fail" "$*")
+	TS_SKIP_LOOPDEVS=$(ts_has_option "skip-loopdevs" "$*")
 
 	tmp=$( ts_has_option "memcheck" "$*")
 	if [ "$tmp" == "yes" -a -f /usr/bin/valgrind ]; then
@@ -248,6 +252,7 @@ function ts_init_env {
 	declare -a TS_SUID_PROGS
 	declare -a TS_SUID_USER
 	declare -a TS_SUID_GROUP
+	declare -a TS_LOOP_DEVS
 
 	if [ -f $TS_TOPDIR/commands.sh ]; then
 		. $TS_TOPDIR/commands.sh
@@ -406,11 +411,7 @@ function ts_finalize_subtest {
 }
 
 function ts_finalize {
-	for idx in $(seq 0 $((${#TS_SUID_PROGS[*]} - 1))); do
-		PROG=${TS_SUID_PROGS[$idx]}
-		chmod a-s $PROG &> /dev/null
-		chown ${TS_SUID_USER[$idx]}.${TS_SUID_GROUP[$idx]} $PROG &> /dev/null
-	done
+	ts_cleanup_on_exit
 
 	if [ $TS_NSUBTESTS -ne 0 ]; then
 		printf "%11s..."
@@ -434,13 +435,23 @@ function ts_finalize {
 
 function ts_die {
 	ts_log "$1"
-	if [ -n "$2" ] && [ -b "$2" ]; then
-		ts_device_deinit "$2"
-		ts_fstab_clean		# for sure...
-	fi
 	ts_finalize
 }
 
+function ts_cleanup_on_exit {
+
+	for idx in $(seq 0 $((${#TS_SUID_PROGS[*]} - 1))); do
+		PROG=${TS_SUID_PROGS[$idx]}
+		chmod a-s $PROG &> /dev/null
+		chown ${TS_SUID_USER[$idx]}.${TS_SUID_GROUP[$idx]} $PROG &> /dev/null
+	done
+
+	for dev in "${TS_LOOP_DEVS[@]}"; do
+		ts_device_deinit "$dev"
+	done
+	unset TS_LOOP_DEVS
+}
+
 function ts_image_md5sum {
 	local img=${1:-"$TS_OUTDIR/${TS_TESTNAME}.img"}
 	echo $(md5sum "$img" | awk '{printf $1}') $(basename "$img")
@@ -455,16 +466,26 @@ function ts_image_init {
 	return 0
 }
 
+function ts_register_loop_device {
+	local ct=${#TS_LOOP_DEVS[*]}
+	TS_LOOP_DEVS[$ct]=$1
+}
+
 function ts_device_init {
 	local img
 	local dev
 
 	img=$(ts_image_init $1 $2)
 	dev=$($TS_CMD_LOSETUP --show -f "$img")
+	if [ "$?" != "0" -o "$dev" = "" ]; then
+		ts_die "Cannot init device"
+	fi
 
-	echo $dev
+	ts_register_loop_device "$dev"
+	TS_LODEV=$dev
 }
 
+# call from ts_cleanup_on_exit() only because of TS_LOOP_DEVS maintenance
 function ts_device_deinit {
 	local DEV="$1"
 
@@ -540,7 +561,7 @@ function ts_is_mounted {
 	grep -q $DEV /proc/mounts && return 0
 
 	if [ "${DEV#/dev/loop/}" != "$DEV" ]; then
-		return grep -q "/dev/loop${DEV#/dev/loop/}" /proc/mounts
+		grep -q "/dev/loop${DEV#/dev/loop/}" /proc/mounts && return 0
 	fi
 	return 1
 }
@@ -609,19 +630,19 @@ function ts_scsi_debug_init {
 	modprobe -r scsi_debug &>/dev/null \
 		|| ts_skip "cannot remove scsi_debug module (rmmod)"
 
-	modprobe -b scsi_debug $* &>/dev/null \
+	modprobe -b scsi_debug "$@" &>/dev/null \
 		|| ts_skip "cannot load scsi_debug module (modprobe)"
 
 	# it might be still not loaded, modprobe.conf or whatever
 	lsmod | grep -q "^scsi_debug " \
 		|| ts_skip "scsi_debug module not loaded (lsmod)"
 
-	devname=$(grep --with-filename scsi_debug /sys/block/*/device/model | awk -F '/' '{print $4}')
-	[ "x${devname}" == "x" ] && ts_die "cannot find scsi_debug device"
-
 	sleep 1
 	udevadm settle
 
+	devname=$(grep --with-filename scsi_debug /sys/block/*/device/model | awk -F '/' '{print $4}')
+	[ "x${devname}" == "x" ] && ts_die "cannot find scsi_debug device"
+
 	TS_DEVICE="/dev/${devname}"
 }
 
@@ -647,3 +668,32 @@ function ts_resolve_host {
 	test -n "$tmp" || return 1
 	echo "$tmp" | sort -R | head -n 1
 }
+
+# listen to unix socket (background socat)
+function ts_init_socket_to_file {
+	local socket=$1
+	local outfile=$2
+	local pid="0"
+
+	ts_check_prog "socat"
+	rm -f "$socket" "$outfile"
+
+	socat -u UNIX-LISTEN:$socket,fork,max-children=1,backlog=128 \
+		STDOUT > "$outfile" &
+	pid=$!
+
+	# check for running background process
+	if [ "$pid" -le "0" ] || ! kill -s 0 "$pid"; then
+		ts_skip "unable to run socat"
+	fi
+	# wait for the socket listener
+	if ! socat -u /dev/null UNIX-CONNECT:$socket,retry=30,interval=0.1; then
+		kill -9 "$pid"
+		ts_skip "timeout waiting for socket"
+	fi
+	# check socket again
+	if ! socat -u /dev/null UNIX-CONNECT:$socket; then
+		kill -9 "$pid"
+		ts_skip "socket stopped listening"
+	fi
+}
